use ExtUtils::MakeMaker;
use ExtUtils::F77;

# check for the modules required by Grid.pm

sub chk_module {
    my($pkg) = @_;
    print "Checking for $pkg ..";
    eval { my $p; ($p = $pkg . ".pm") =~ s#::#/#g; require $p; };
    if ($@) {
	print " failed\n";
	$missing_modules++;
    } else {
	print " ok\n";
    }
}

chk_module('DBI');
chk_module('File::PathConvert');
chk_module('Statistics::Descriptive');

if ($missing_modules) {
    print "One or more required modules is missing. Stop.\n";
    exit;
}

my @libdirs = ('/usr/lib','/usr/local/lib','/usr/local/pgplot','/usr/X11R6/lib');
my @incdirs = ('/usr/include','/usr/local/include','/usr/local/pgplot');

sub fe {
    my($fn,@dirs) = @_;
    foreach (@dirs) {
	return 1 if -e "$_/$fn";
    }
    return 0;
}

print "Checking for PGPLOT ..";
if (fe('libcpgplot.a',@libdirs) and fe('libpgplot.a',@libdirs) and 
    fe('cpgplot.h',@incdirs)) {
    print " ok\n";
} else {
    print " failed\n";
}
print "Checking for libgr .."; # 
if (fe('libpbm.so',@libdirs) and fe('libpgm.so',@libdirs) and 
    fe('libppm.so',@libdirs) and fe('ppm.h',@incdirs)) {
    print " ok\n";
} else {
    print " failed\n";
}

#todo: check for pgplot & ppm

my $libs = '';
foreach (@libdirs) {
    $libs .= " -L$_";
}

# See lib/ExtUtils/MakeMaker.pm for details of how to influence
# the contents of the Makefile that is written.
WriteMakefile(
	      'NAME'	     => 'Grid',
	      'VERSION_FROM' => 'Grid.pm',
	      'LIBS'	     => ["$libs -lcpgplot -lpgplot -lpbm -lpgm -lppm -lX11 " . 
			       ExtUtils::F77->runtime],
	      'DEFINE'	     => '-DHAVE_PPM -DHAVE_PGPLOT',
	      'INC'	     => '',
	      'MYEXTLIB'     => 'grid/libgrid$(LIB_EXT)'
	      );

sub MY::postamble {
'
cvsupdate:
	cvs -d :pserver:anonymous@kampela.hut.fi:/var/cvsroot -z3 update

cvscommit:
	cvs -d :pserver:ajolma@kampela.hut.fi:/var/cvsroot -z3 commit

$(MYEXTLIB): grid/Makefile
	cd grid && $(MAKE) $(PASTHRU)
';
}
