#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "grid.h"
#include "graphics.h"

#ifdef HAVE_PGPLOT
#include <cpgplot.h>
void cpgpt1(float xpt, float ypt, int symbol); /* WHY THIS ??? */
#endif


color_table *ctcreate(int nc, double contra, double bright)
{
    static char *fct = "ctcreate";
    color_table *ct = (color_table *)malloc(sizeof(color_table));
    if (!ct) {
	fprintf(stderr,"%s: Out of memory!\n", fct);
	return NULL;
    }
    ct->l = NULL;
    ct->r = NULL;
    ct->g = NULL;
    ct->b = NULL;
    ct->nc = nc;
    ct->contra = contra;
    ct->bright = bright;
    if (nc == 0) return ct;
    if (!(ct->l = (float *)calloc(nc,sizeof(float)))) goto fail;
    if (!(ct->r = (float *)calloc(nc,sizeof(float)))) goto fail;
    if (!(ct->g = (float *)calloc(nc,sizeof(float)))) goto fail;
    if (!(ct->b = (float *)calloc(nc,sizeof(float)))) goto fail;
    return ct;
 fail:
    fprintf(stderr,"%s: Out of memory!\n", fct);
    ctdestroy(ct);
    return NULL;
}


int ctdestroy(color_table *ct)
{
    if (ct) {
	if (ct->l) free(ct->l);
	if (ct->r) free(ct->r);
	if (ct->g) free(ct->g);
	if (ct->b) free(ct->b);
	free(ct);
    }
    return 1;
}


int ctset(color_table *ct, int i, double l, double r, double g, double b)
{
    static char *fct = "ctset";
    if (!ct OR i < 0 OR i >= ct->nc) {
	fprintf(stderr,"%s: %i: color table index out of bounds\n",fct,i);
	return 0;
    }
    ct->l[i] = l;
    ct->r[i] = r;
    ct->g[i] = g;
    ct->b[i] = b;
    return 1;
}

int ctsize(color_table *ct) {
    if (ct) return ct->nc;
    return 0;
}


double ctget_contrast(color_table *ct)
{
    return ct->contra;
}


double ctget_brightness(color_table *ct)
{
    return ct->bright;
}


double ctgetl(color_table *ct, int i) {
    static char *fct = "ctgetl";
    if (!ct OR i < 0 OR i >= ct->nc) {
	fprintf(stderr,"%s: %i: color table index out of bounds\n",fct,i);
	return 0;
    }
    return ct->l[i];
}


double ctgetr(color_table *ct, int i) {
    static char *fct = "ctgetr";
    if (!ct OR i < 0 OR i >= ct->nc) {
	fprintf(stderr,"%s: %i: color table index out of bounds\n",fct,i);
	return 0;
    }
    return ct->r[i];
}


double ctgetg(color_table *ct, int i) {
    static char *fct = "ctgetg";
    if (!ct OR i < 0 OR i >= ct->nc) {
	fprintf(stderr,"%s: %i: color table index out of bounds\n",fct,i);
	return 0;
    }
    return ct->g[i];
}


double ctgetb(color_table *ct, int i) {
    static char *fct = "ctgetb";
    if (!ct OR i < 0 OR i >= ct->nc) {
	fprintf(stderr,"%s: %i: color table index out of bounds\n",fct,i);
	return 0;
    }
    return ct->b[i];
}


#ifdef HAVE_PGPLOT

int scale(const char *name, xyrange *xy, double *x, double *y)
{
    if ((xy->minX)>0) {
	/* scale the axis */
	(*x) = pow(10,floor(log10((xy->minX))));
	if ((*x)>1.001 OR (*x)<0.999) {
	    (xy->minX) /= (*x);
	    (xy->maxX) /= (*x);
	    fprintf(stderr,"%s: x values are divided by %f\n",name,(*x));
	} else {
	    (*x) = 1;
	}
    } 
    if ((xy->minY)>0) {
	(*y) = pow(10,floor(log10((xy->minY))));
	if ((*y)>1.001 OR (*y)<0.999) {
	    (xy->minY) /= (*y);
	    (xy->maxY) /= (*y);
	    fprintf(stderr,"%s: y values are divided by %f\n",name,(*y));
	} else {
	    (*y) = 1;
	}
    }
    return 1;
}


void settr(float *tr, int M, int N, xyrange xy)
{
    double dx = (xy.maxX-xy.minX)/N;
    double dy = (xy.maxY-xy.minY)/M;
    tr[0]=xy.minX-dx/2;
    tr[1]=(xy.maxX-xy.minX)/N;
    tr[2]=0;
    tr[3]=xy.maxY+dy/2;
    tr[4]=0;
    tr[5]=-(xy.maxY-xy.minY)/M;
}


int cpg(grid *gd, vector_data *vd, color_table *ct, float *width, float *tr, xyrange xy, int options, double sx, double sy)
{
    double aspect = (xy.maxY-xy.minY)/(xy.maxX-xy.minX);
    float x1, x2, y1, y2;
    cpgpap(*width, aspect);
    cpgqvsz(1, &x1, &x2, &y1, &y2);
    *width = x2;
    cpgwnad(xy.minX,xy.maxX+(xy.maxX-xy.minX)/100,
	    xy.minY,xy.maxY+(xy.maxY-xy.minY)/100);
    cpgbox("BCINTS", 0, 0, "BCINTS", 0, 0);
    /* grid i,j are not pgplot i,j! */
    if (ct AND ct->nc > 0) { 
	/* 
	   it is assumed here that the colortable has 
	   at least the size maxval - minval + 1 
	   the extra colorspace is above the image colors
	*/
	cpgscir(0, ct->nc-1);
	cpgctab(ct->l, ct->r, ct->g, ct->b, ct->nc, ct->contra, ct->bright);
	if (options & RASTER_FLAG) 
	    cpgimag(gd->fdata, gd->N, gd->M, 1, gd->N, 1, gd->M, 
		    gd->minval, gd->minval + ct->nc - 1, tr);
    } else {
	cpgscir(0,255);
	if (options & RASTER_FLAG) 
	    cpggray(gd->fdata, gd->N, gd->M, 1, gd->N, 1, gd->M, 
		    gd->maxval, gd->minval, tr);
    }
    if (vd AND vd->id AND (options & VECTOR_FLAG)) {
	while (vd) {    
	    if (vd->n_points) {	
		int i;
		for (i = 0; i < vd->n_points; i++) {
		    point_data *p = vd->pointtable[i];
		    cpgsci(p->color);
		    cpgslw(p->line_width);
		    if (p->symbol > 31) cpgscf(p->symbol_font);
		    cpgsch(p->symbol_height);
		    cpgpt1(p->p.x/sx, p->p.y/sy, p->symbol);
		}
	    }
	    if (vd->n_lsegs) {
		int i;
		for (i = 0; i < vd->n_lsegs; i++) {
		    lseg_data *l = vd->lsegtable[i];
		    cpgsls(l->style);
		    cpgslw(l->width);
		    cpgsci(l->color);
		    if (l->arrow) {
			cpgsah(l->arrow, l->arrow_angle, l->arrow_barb);
			cpgarro(l->l.p[0].x/sx, l->l.p[0].y/sy, 
				l->l.p[1].x/sx, l->l.p[1].y/sy);
		    } else {
			float x[2],y[2];
			x[0] = l->l.p[0].x/sx;
			x[1] = l->l.p[1].x/sx;
			y[0] = l->l.p[0].y/sy;
			y[1] = l->l.p[1].y/sy;
			cpgline(2, x, y);
		    }
		}
	    }
	    if (vd->n_polygons) {
		int i;
		for (i = 0; i < vd->n_polygons; i++) {		    
		    polygon_data *p = vd->polygontable[i];
		    float *x;
		    float *y;
		    int j;
		    cpgsfs(p->fill_style);
		    if (p->fill_style > 2) /* hatched */
			cpgshs(p->fill_angle, p->fill_sepn, p->fill_phase);
		    cpgsls(p->line_style);
		    cpgslw(p->line_width);
		    cpgsci(p->color);
		    x = calloc(p->n, sizeof(float));
		    if (x) {
			y = calloc(p->n, sizeof(float));
			if (y) {
			    for (j = 0; j < p->n; j++) {
				x[j] = p->points[j].x/sx;
				y[j] = p->points[j].y/sy;
			    }
			    cpgline(p->n, x, y);
			    free(y);
			}
			free(x);
		    }
		}
	    }
	    /* labels */
	    if (vd->id AND options & LABEL_FLAG) {
		cpgsci(vd->label_color);
		cpgslw(vd->label_line_width);
		cpgscf(vd->label_font);
		cpgsch(vd->label_height);
		if (vd->n_points) {	
		    int i;
		    for (i = 0; i < vd->n_points; i++) {
			point_data *p = vd->pointtable[i];
			if (p->label AND *(p->label)) {
			    /* this should depend on the label placement */
			    float x = p->p.x/sx;
			    float y = p->p.y/sy;
			    if (x >= xy.minX AND x <= xy.maxX AND 
				y >= xy.minY AND y <= xy.maxY)
				cpgptxt(x, y, 0, 0, p->label);
			}
		    }
		}
		if (vd->n_lsegs) {
		    int i;
		    for (i = 0; i < vd->n_lsegs; i++) {
			lseg_data *l = vd->lsegtable[i];
			if (l->label AND *(l->label)) {
			    /* this should depend on the label placement */
			    float x = ((l->l.p[0].x+l->l.p[1].x)/2)/sx;
			    float y = ((l->l.p[0].y+l->l.p[1].y)/2)/sy;
			    if (x >= xy.minX AND x <= xy.maxX AND 
				y >= xy.minY AND y <= xy.maxY)
				cpgptxt(x, y, 0, 0.5, l->label);
			}
		    }
		}
		if (vd->n_polygons) {
		    int i;
		    for (i = 0; i < vd->n_polygons; i++) {
			polygon_data *p = vd->polygontable[i];
			if (p->label AND *(p->label)) {
			    /* this should depend on the label placement */
			    float xmin = p->points[0].x;
			    float xmax = xmin; 
			    float ymin = p->points[0].y;
			    float ymax = ymin;
			    int j;
			    for (j = 1; j < p->n; j++) {
				if (p->points[j].x < xmin) 
				    xmin = p->points[j].x;
				if (p->points[j].x > xmax) 
				    xmax = p->points[j].x;
				if (p->points[j].y < ymin) 
				    ymin = p->points[j].y;
				if (p->points[j].y > ymax) 
				    ymax = p->points[j].y;
			    }
			    {
				float x = (xmin+xmax)/2/sx;
				float y = (ymin+ymax)/2/sy;
				if (x >= xy.minX AND x <= xy.maxX AND 
				    y >= xy.minY AND y <= xy.maxY)
				    cpgptxt(x, y, 0, 0.5, p->label);
			    }
			}
		    }
		}
	    }
	    vd = vd->next;
	}
    }
    return 1;
}


void painthelp(char *device, int draw, int options)
{
    device++;
    printf("keys: <q> quit, <e> export as %s, <f> redraw, <i> toggle info, <h> help\n", device);
    if (options & RASTER_FLAG)
	printf("      <R> toggle raster off, ");
    else
	printf("      <R> toggle raster on,  ");
    if (options & VECTOR_FLAG)
	printf("<V> toggle vector data off, ");
    else
	printf("<V> toggle vector data on,  ");
    if (options & LABEL_FLAG)
	printf("<L> toggle labels off\n");
    else
	printf("<L> toggle labels on\n");
    printf("      <2,4,6,8> scroll, <+-> zoom in/out, <a> calculate area\n");
    if (draw) printf("modes: <z> zoom, <l> line, <r> filled rectangle, <c> filled circle\n");
}


void modehelp(char mode)
{
    switch (mode) {
    case 'c':
	printf("draw a filled circle: use left button to draw, right button cancels\n");
	break;
    case 'r':
	printf("draw a filled rectangle: use left button to draw, right button cancels\n");
	break;
    case 'l':
	printf("draw a line: use left button to draw, right button cancels\n");
	break;
    case 'z':
	printf("zoom mode: use left button to zoom, right button cancels/removes zoom\n");
	break;
    }
}


int gdwindow_open()
{
    int window;
    window = cpgopen("/xserve");
    return window;
}


int gdwindow_close(int window)
{
    if (window < 1) return 0;
    cpgslct(window);
    cpgclos();
    return 0;
}


/* the default width of the window in inches */

#define DEFAULT_WIDTH 7

grid *gdpaint(grid *gd, 
	      vector_data *vd, 
	      char *device,
	      int window,
	      color_table *ct, 
	      int draw, 
	      int options 
    )
{
    const char *fct = "gdpaint";

    /* 
       "cdevice" is the primary device        
       which is /xserve always if draw >= 0 
       "device" is the export device in that case
    */

    char *cdevice = device;
    grid *ng = NULL, *dg = NULL;
    int destroy = 0, info = 0;
    xyrange xy = gd->xy, view;
    double x_scale = 1, y_scale = 1;
    float tr[6], width;
    if ((gd->xy.maxX <= gd->xy.minX) OR (gd->xy.maxY <= gd->xy.minY)) {
	fprintf(stderr,"%s: min{X|Y} of grid >= max{X|Y} of grid\n", fct);
	goto fail0;
    }
    if (gd->minval >= gd->maxval) gdsetminmax(gd);
    if (gd->minval >= gd->maxval) {
	fprintf(stderr,"%s: min of grid >= max of grid\n", fct);
	goto fail0;
    }
    if (gd->datatype != REAL_GRID) {
	if (!(gd = gdcreatecopy(gd, REAL_GRID))) goto fail0;
	destroy = 1;
    }
    if (draw > 0) {
	if (!(ng = gdnew2(INTEGER_GRID, gd))) goto fail0;
	if (!(dg = gdcreatecopy(gd, REAL_GRID))) goto fail0;
	dg->minval = gd->minval;
	dg->maxval = gd->maxval + 1; /* 1 for drawing */
    } else
	dg = gd;
    if (options & SCALE_FLAG) scale(fct, &xy, &x_scale, &y_scale);
    settr(tr, dg->M, dg->N, xy);
    view = xy;
    width = DEFAULT_WIDTH;

    if (draw >= 0) cdevice = "/xserve";
    if (window < 1) {
	window = cpgopen(cdevice);
	if (window < 1) goto fail;
    }
    cpgslct(window);
    cpg(dg, vd, ct, &width, tr, view, options, x_scale, y_scale);

    if (draw >= 0) {
	char key = 'x', mode = 'z';
	double dxy;
	/*int color = dg->maxval;*/
	painthelp(device, draw, options);
	modehelp(mode);
	while (key != 'q') {
	    xyrange band;
	    cell c1 = {0, 0}, c2 = {0, 0};
	    cpgcurs(&band.minX, &band.minY, &key);
	    if (!strchr("Xirhq", key)) {
		c1.i = gdy2i(dg, y_scale * band.minY);
		c1.j = gdx2j(dg, x_scale * band.minX);
		if (info) 
		    printf("(x,y)=(%f,%f) (i,j)=(%i,%i) val=%.4f %c\n",
			   band.minX, band.minY, c1.i, c1.j, GDGET(gd, c1), key);
	    }
	    width = 0;
	    switch (key) {
	    case 'R':
		if (options & RASTER_FLAG)
		    options &= (VECTOR_FLAG + LABEL_FLAG + SCALE_FLAG);
		else
		    options |= RASTER_FLAG;
		break;
	    case 'V':
		if (options & VECTOR_FLAG)
		    options &= (RASTER_FLAG + LABEL_FLAG + SCALE_FLAG);
		else
		    options |= VECTOR_FLAG;
		break;
	    case 'L':
		if (options & LABEL_FLAG)
		    options &= (RASTER_FLAG + VECTOR_FLAG + SCALE_FLAG);
		else
		    options |= LABEL_FLAG;
		break;	    
	    case 'A':
		switch (mode) {
		case 'l':
		    cpgband(1, 0, band.minX, band.minY, &band.maxX, &band.maxY, &key);
		    if (key == 'A') {
			c2.i = gdy2i(dg, y_scale * band.maxY);
			c2.j = gdx2j(dg, x_scale * band.maxX);
			gdline(ng, c1.i, c1.j, c2.i, c2.j, 1);
			gdline(dg, c1.i, c1.j, c2.i, c2.j, dg->maxval);
			band.minX = band.maxX;
			band.minY = band.maxY;
		    }
		    break;
		case 'r':
		    cpgband(2, 0, band.minX, band.minY, &band.maxX, &band.maxY, &key);
		    if (key == 'A') {
			c2.i = gdy2i(gd, y_scale * band.maxY);
			c2.j = gdx2j(gd, x_scale * band.maxX);
			gdfilledrect(ng, c1.i, c1.j, c2.i, c2.j, 1);
			gdfilledrect(dg, c1.i, c1.j, c2.i, c2.j, dg->maxval);
		    }
		    break;
		case 'c':
		    cpgband(1, 0, band.minX, band.minY, &band.maxX, &band.maxY, &key);
		    if (key == 'A') {
			double r2;
			c2.i = gdy2i(gd, y_scale * band.maxY);
			c2.j = gdx2j(gd, x_scale * band.maxX);
			r2 = (c2.i - c1.i)*(c2.i - c1.i) + (c2.j - c1.j)*(c2.j - c1.j);
			gdfilledcircle(ng, c1.i, c1.j, round(sqrt(r2)), round(r2), 1);
			gdfilledcircle(dg, c1.i, c1.j, round(sqrt(r2)), round(r2), dg->maxval);
		    }
		    break;
		case 'z':
		    cpgband(2, 0, band.minX, band.minY, &band.maxX, &band.maxY, &key);
		    if (key == 'A') {
			int c;
			c2.i = gdy2i(gd, y_scale * band.maxY);
			c2.j = gdx2j(gd, x_scale * band.maxX);
			if (c2.i < c1.i) swap(c1.i,c2.i,c);
			if (c2.j < c1.j) swap(c1.j,c2.j,c);
			view = band;
			if (band.maxX < band.minX) {
			    view.minX = band.maxX;
			    view.maxX = band.minX;
			}
			if (band.maxY < band.minY) {
			    view.minY = band.maxY;
			    view.maxY = band.minY;
			}
			if (view.maxY > view.minY AND view.maxX > view.minX AND info)
			    printf("(x,y)=(%f,%f) (i,j)=(%i,%i)\n", band.maxX, band.maxY, c2.i, c2.j);
		    }
		    break;
		}
		break;
	    case 'X':
		view = xy;
		width = DEFAULT_WIDTH;
		break;
	    case 'a':
		{
		    double size = gdzonesize(gd, c1);
		    printf("(x,y)=(%f,%f) (i,j)=(%i,%i) val=%.4f size=%f\n",
			   band.minX, band.minY, c1.i, c1.j, gdget(gd, c1.i, c1.j), size);
		    break;
		}
	    case 'i':
		info = !info;
		if (info)
		    printf("info on\n");
		else
		    printf("info off\n");
		break;
	    case 'c':
	    case 'r':
	    case 'l':
		if (!draw) break;
	    case 'z':
		mode = key;
		modehelp(mode);
		break;
	    case '+':
		dxy = (view.maxY - view.minY)/4;
		view.minY += dxy;
		view.maxY -= dxy;
		dxy = (view.maxX - view.minX)/4;
		view.minX += dxy;
		view.maxX -= dxy;
		break;
	    case '-':
		dxy = (view.maxY - view.minY)/4;
		view.minY -= dxy;
		view.maxY += dxy;
		dxy = (view.maxX - view.minX)/4;
		view.minX -= dxy;
		view.maxX += dxy;
		break;
	    case '2':
		dxy = (view.maxY - view.minY)/2;
		view.minY -= dxy;
		view.maxY -= dxy;
		break;
	    case '4':
		dxy = (view.maxX - view.minX)/2;
		view.minX -= dxy;
		view.maxX -= dxy;
		break;
	    case '6':
		dxy = (view.maxX - view.minX)/2;
		view.minX += dxy;
		view.maxX += dxy;
		break;
	    case '8':
		dxy = (view.maxY - view.minY)/2;
		view.minY += dxy;
		view.maxY += dxy;
		break;
	    case 'e': {
		int eid = cpgopen(device);
		cpgslct(eid);
		cpg(dg, vd, ct, &width, tr, view, options, x_scale, y_scale);
		cpgclos();
		width = 0;
	    }
	    break;
	    case 'h':
		painthelp(device, draw, options);
	    }
	    if (strchr("RVLAX+-2468ef", key)) {
		cpgslct(window);
		cpgclos();
		window = cpgopen(cdevice);
		if (window < 1) goto fail;
		cpgslct(window);
		cpg(dg, vd, ct, &width, tr, view, options, x_scale, y_scale);
	    }
	}
    }
    goto ok;
 fail0:
    window = 0;
 fail:
    if (ng) gddestroy(ng);
    ng = NULL;
 ok:
    if (window) {
	cpgslct(window);
	cpgclos();
    }
    if (destroy) gddestroy(gd);
    if (draw > 0) gddestroy(dg);
    return ng;
}

#endif

