#include <stdio.h>
#include <stdlib.h>

#include "hash.h"

/*
** BASED ON
** public domain code by Jerry Coffin, with improvements by HenkJan Wolthuis.
*/

static unsigned hashfct(int key)
{
    unsigned ret_val = 0;
    ret_val ^= key;
    ret_val <<= 1;
    return ret_val;
}


int hash_create(hash *table, size_t size)
{
    table->size  = size;
    table->table = (hash_item **)calloc(size, sizeof(void *));
    if (!table->table) {
	table->size  = 0;
	fprintf(stderr,"Out of memory!\n");
	return 0;
    }
    return 1;
}


int hash_del(hash *table, int key)
{
    unsigned val = hashfct(key) % table->size;
    hash_item *ptr, *last;
    
    if (!(table->table)[val])
	return 0;

    for (last = NULL, ptr = (table->table)[val]; ptr; last = ptr, ptr = ptr->next) {

	if (key == ptr->key) {
	    if (last) {
		last->next = ptr->next;
		free(ptr);
		return 1;
	    } else {
		(table->table)[val] = ptr->next;
		free(ptr);
		return 1;
	    }
	}
    }
    return 0;
}


void hash_destroy(hash *table)
{
    unsigned i;
    for (i = 0; i < table->size; i++) {
	hash_item *a = table->table[i];
	while (a) {
	    hash_item *b = a->next;
	    free(a);
	    a = b;
	}
    }
    free(table->table);
    table->table = NULL;
    table->size = 0;
}


int hash_count(hash *table)
{
    unsigned count = 0, i;
    for (i = 0; i < table->size; i++) {
	hash_item *a = table->table[i];
	while (a) {
	    count++;
	    a = a->next;
	}
    }
    return count;
}


int *hash_keys(hash *table, int *count)
{
    int *keys, i;
    *count = 0;
    if (!(keys = (int *)calloc(hash_count(table), sizeof(int)))) {
	fprintf(stderr,"Out of memory!\n");
	return NULL;
    }
    for (i = 0; i < table->size; i++) {
	hash_item *a = table->table[i];
	while (a) {
	    keys[*count] = a->key;
	    (*count)++;
	    a = a->next;
	}
    }
    return keys;
}


int hash_int_insert(hash *table, int key, int value)
{
    unsigned val = hashfct(key) % table->size;
    hash_int_item *ptr;

    if (!(table->table)[val]) {
	ptr = malloc(sizeof(hash_int_item));
	if (!ptr) {
	    fprintf(stderr,"Out of memory!\n");
	    return 0;
	}	
	ptr->key = key;
	ptr->value = value;
	ptr->next = NULL;
	(table->table)[val] = (hash_item *)ptr;
    }

    for (ptr = (hash_int_item *)(table->table)[val]; ptr; ptr = ptr->next)
	if (key == ptr->key) {
	    ptr->value = value;
	    return 1;
	}

    ptr = (hash_int_item *)malloc(sizeof(hash_int_item));
    if (!ptr) {
	fprintf(stderr,"Out of memory!\n");
	return 0;
    }
    ptr->key = key;
    ptr->value = value;
    ptr->next = (hash_int_item *)(table->table)[val];
    (table->table)[val] = (hash_item *)ptr;
    return 1;
}

int hash_int_lookup(hash *table, int key, int **value)
{
    unsigned val = hashfct(key) % table->size;
    hash_int_item *ptr;

    if (!(table->table)[val])
	return 0;

    for (ptr = (hash_int_item *)(table->table)[val]; ptr; ptr = ptr->next) {
	if (key == ptr->key ) {
	    *value = &(ptr->value);
	    return 1;
	}
    }
    return 0;
}

void hash_int_enumerate(hash *table, void (*func)(int, int))
{
    unsigned i;
    hash_int_item *temp;

    for (i = 0; i < table->size; i++) {
	if ((table->table)[i]) {
	    for (temp = (hash_int_item *)(table->table)[i];
		 NULL != temp;
		 temp = temp->next)
		{
		    func(temp->key, temp->value);
		}
	}
    }
}


int hash_double_insert(hash *table, int key, double value)
{
    unsigned val = hashfct(key) % table->size;
    hash_double_item *ptr;

    if (!(table->table)[val]) {
	ptr = malloc(sizeof(hash_double_item));
	if (!ptr) {
	    fprintf(stderr,"Out of memory!\n");
	    return 0;
	}	
	ptr->key = key;
	ptr->value = value;
	ptr->next = NULL;
	(table->table)[val] = (hash_item *)ptr;
    }

    for (ptr = (hash_double_item *)(table->table)[val]; ptr; ptr = ptr->next)
	if (key == ptr->key) {
	    ptr->value = value;
	    return 1;
	}

    ptr = (hash_double_item *)malloc(sizeof(hash_double_item));
    if (!ptr) {
	fprintf(stderr,"Out of memory!\n");
	return 0;
    }
    ptr->key = key;
    ptr->value = value;
    ptr->next = (hash_double_item *)(table->table)[val];
    (table->table)[val] = (hash_item *)ptr;
    return 1;
}

int hash_double_lookup(hash *table, int key, double **value)
{
    unsigned val = hashfct(key) % table->size;
    hash_double_item *ptr;

    if (!(table->table)[val])
	return 0;

    for (ptr = (hash_double_item *)(table->table)[val]; ptr; ptr = ptr->next) {
	if (key == ptr->key ) {
	    *value = &(ptr->value);
	    return 1;
	}
    }
    return 0;
}

void hash_double_enumerate(hash *table, void (*func)(int, double))
{
    unsigned i;
    hash_double_item *temp;

    for (i = 0; i < table->size; i++) {
	if ((table->table)[i]) {
	    for (temp = (hash_double_item *)(table->table)[i];
		 NULL != temp;
		 temp = temp->next)
		{
		    func(temp->key, temp->value);
		}
	}
    }
}


