#ifndef HASH__H
#define HASH__H

#include <stddef.h>

typedef struct hash_item {
    int key;
    struct hash_item *next;
} hash_item;

typedef struct hash_int_item {
    int key;
    struct hash_int_item *next;
    int value;
} hash_int_item;

typedef struct hash_double_item {
    int key;
    struct hash_double_item *next;
    double value;
} hash_double_item;

typedef struct {
    size_t size;
    hash_item **table;
} hash;

int hash_create(hash *table, size_t size);
int hash_del(hash *table, int key);
void hash_destroy(hash *table);
int hash_count(hash *table);
int *hash_keys(hash *table, int *count);

int hash_int_insert(hash *table, int key, int value);
int hash_int_lookup(hash *table, int key, int **value);
void hash_int_enumerate(hash *table, void (*func)(int, int));

int hash_double_insert(hash *table, int key, double value);
int hash_double_lookup(hash *table, int key, double **value);
void hash_double_enumerate(hash *table, void (*func)(int, double));

#endif /* HASH__H */
